import pandas as pd
import black
import json


with open("intel_cpu_database.json", "r") as handle:
    intel_raw = json.loads(handle.read())

with open("AMD_data.json", "r") as handle:
    amd_raw = json.loads(handle.read())
    amd_raw = pd.DataFrame(amd_raw["data"])


### Parse AMD


def parse_amd_clock(name):
    name = name.strip().split()

    if len(name) == 0:
        return None

    name = name[-1]

    if "mhz" in name.lower():
        repl = "mhz"
        coef = 1e6

    elif "ghz" in name.lower():
        repl = "ghz"
        coef = 1e9
    else:
        raise KeyError

    return int(coef * float(name.lower().replace(repl, "").strip()))


def parse_amd_launch(d):

    if d is None:
        return None

    if len(d) == 0:
        return None

    try:
        return int(d.split("/")[-1])
    except ValueError:
        if (d[0] == "Q") or (d[1] == "Q"):
            return int(d[2:].strip())
        if "September" in d:
            return int(d.split()[-1])
        print(d)


amd = pd.DataFrame(index=amd_raw.index)
amd["ncores"] = amd_raw["# of CPU Cores"]
amd["nthreads"] = amd_raw["# of Threads"]
amd["model"] = amd_raw["Model"]
amd["family"] = amd_raw["Family"]
amd["base_clock"] = amd_raw["Base Clock"].apply(parse_amd_clock)
amd["boost_clock"] = amd_raw["Max Boost Clock"].apply(parse_amd_clock)
amd["launch_date"] = amd_raw["Launch Date"].apply(parse_amd_launch)
amd["target_use"] = amd_raw["Platform"]
amd["vendor"] = "amd"

amd["microarchitecture"] = None
amd.loc[amd_raw["Supported Technologies"].str.contains("Zen "), "microarchitecture"] = "Zen"
amd.loc[amd_raw["Supported Technologies"].str.contains("Zen\+"), "microarchitecture"] = "Zen+"
amd.loc[amd_raw["Supported Technologies"].str.contains("Zen\+\+"), "microarchitecture"] = "Zen++"

amd["type"] = "cpu"
# amd.loc[amd['model'].str.contains("APU"), "type"] = "apu"

amd["instructions"] = None

# Each of these families have at least this instruction set
amd.loc[amd_raw["Family"].str.contains("Athlon"), "instructions"] = "sse"
amd.loc[amd_raw["Family"].str.contains("Phenom"), "instructions"] = "sse"
amd.loc[amd_raw["Family"].str.contains("FX-Series"), "instructions"] = "sse"
amd.loc[amd_raw["Family"].str.contains("Opteron"), "instructions"] = "sse"
amd.loc[amd_raw["Family"].str.contains("EPYC"), "instructions"] = "avx2"
amd.loc[amd_raw["Family"].str.contains("Ryzen"), "instructions"] = "avx2"

# Climb the ladder
amd.loc[amd_raw["Supported Technologies"].str.contains("SSE"), "instructions"] = "sse"
amd.loc[amd_raw["Supported Technologies"].str.contains("AVX"), "instructions"] = "avx"
amd.loc[amd_raw["Supported Technologies"].str.contains("AVX2"), "instructions"] = "avx2"

amd.loc[amd_raw["Supported Technologies"].str.contains("Zen"), "instructions"] = "avx2"


### Intel Data

intel_rows = []
for row in list(intel_raw.values())[100:]:
    # #     for k, v in row.items():
    # #         print(k)
    # #         print(v)
    # #         print()

    # Not a processor
    if "Performance" not in row:
        continue

    # Some processors are classified as "Mobile", which are almost certainly in laptops not phones
    # Excluding atom procs instead, since they're the ones the break the parser
    #if row["Essentials"]["Vertical Segment"] == "Mobile":
    #    continue
    if "Intel Atom " in row["name"]:
        continue

    try:
        proc = {
            "ncores": int(row["Performance"]["# of Cores"]),
            "nthreads": row["Performance"].get("# of Threads", None),
            "base_clock": row["Performance"]["Processor Base Frequency"],
            "boost_clock": row["Performance"].get("Max Turbo Frequency", None),
            "model": row["Essentials"].get("Processor Number", None),
            "family": row["Essentials"]["Product Collection"],
            "launch_date": row["Essentials"].get("Launch Date"),
            "target_use": row["Essentials"]["Vertical Segment"],
            "vendor": "intel",
            "microarchitecture": row["Essentials"].get("Code Name"),
            "instructions": row.get("Advanced Technologies", {}).get("Instruction Set Extensions"),
        }
        intel_rows.append(proc)
    except:
        print(row)
        raise



def parse_intel_clock(name):

    if name is None:
        return None

    name = name.lower()


    if "mhz" in name:
        repl = "mhz"
        coef = 1e6

    elif "ghz" in name:
        repl = "ghz"
        coef = 1e9
    else:
        raise KeyError

    name = name.replace(repl, "").strip()

    return int(coef * float(name))


def parse_instructions(inst):

    if inst is None:
        return None

    for i in ["AVX-512", "AVX2", "AVX", "SSE"]:
        if i in inst:
            return i

    return None


def parse_date(d):

    if d is None:
        return None

    month, year = d.split("'")

    return int(year) + 2000


intel = pd.DataFrame(intel_rows)
intel["base_clock"] = intel["base_clock"].apply(parse_intel_clock)
intel["boost_clock"] = intel["boost_clock"].apply(parse_intel_clock)
intel["microarchitecture"] = [x.replace("Products formerly ", "") if x else x for x in intel["microarchitecture"]]
intel["instructions"] = intel["instructions"].apply(parse_instructions)
intel["launch_date"] = intel["launch_date"].apply(parse_date)
intel["type"] = "cpu"

df = pd.concat([amd, intel], sort=True)
df.dropna(how="all", inplace=True)

# Munge instructions
df["instructions"] = df["instructions"].str.lower()
df.loc[df["instructions"] == "avx-512", "instructions"] = "avx512"
df.loc[df["instructions"].isnull(), "instructions"] = "none"

translation = {"none": 0, "sse": 1, "avx": 2, "avx2": 3, "avx512": 4}
df["instructions"] = df["instructions"].apply(lambda x: translation[x])


# add extra data
import extra_cpus
for i in extra_cpus.extra_cpus:
    df = df.append(i, ignore_index=True)

def name(vendor, family, model, clock_speed):
    if vendor.lower() in family.lower():
        vendor = ""
    if family in str(model):
        family = ""
    if family.endswith("Processors") and family[:-len("Processors")] in str(model):
        family = ""

    return f"{vendor} {family} {model} @ {clock_speed/1_000_000_000:.1f} GHz"

df["name"] = df.apply(lambda row: name(row["vendor"], row["family"], row["model"], row["base_clock"]), axis=1)

for (vendor, model), fix in extra_cpus.fixes.items():
    idx = df[(df["vendor"] == vendor) & (df["model"] == model)].index
    for k, v in fix.items():
        df.loc[idx, k] = v
# Print some data for posterity
print(df[df["vendor"] == "intel"].tail())
print(df[df["vendor"] == "amd"].tail())
print('---')

# Handle nthreads == ncore bugs
mask = (df["nthreads"] == "") | df["nthreads"].isnull()
df.loc[mask, "nthreads"] = df.loc[mask, "ncores"]

mask = (df["nthreads"] != "") & df["nthreads"].notnull()
#print(df[~mask])
cnt = df.shape[0]
df = df[mask]
print(f"Dropped {cnt - df.shape[0]} / {cnt} processors without ncores")

# Strip out bad models

cnt = df.shape[0]
df = df[~(df["model"].isnull() & (df["launch_date"].isnull() | df["launch_date"] < 2008))]
print(f"Dropped {cnt - df.shape[0]} / {cnt} processors without model numers")


df.sort_values(["vendor", "model", "launch_date"], inplace=True)
df.drop_duplicates(subset=["vendor", "model"], keep="last", inplace=True)



output = f'''
"""
Processor data from multiple sources and vendors.

File Authors: QCElemental Authors
"""

'''

def to_python_str(data):
    return json.dumps(data, indent=2).replace("true", "True").replace("false", "False").replace("NaN", "None").replace("null", "None")

output += f"data_rows = {to_python_str([tuple(x[1].values) for x in df.iterrows()])}\n" 
output += f"data_columns = {to_python_str(list(df.columns))}\n"
output += "data_blob = [{k: v for k, v in zip(data_columns, row)} for row in data_rows]\n"

output = black.format_str(output, mode=black.FileMode())

fn = "cpu_data_blob.py"
with open(fn, "w") as handle:
    handle.write(output)



"""
This file will generate a JSON blob usable by QCElemental for physical constants
"""

import json
import datetime
import requests
from yapf.yapflib.yapf_api import FormatCode

metadata_file = "../../raw_data/srd121_nist-codata-fundamental-physical-constants-2014-metadata.json"
with open(metadata_file, "r") as handle:
    metadata = json.load(handle)

title = metadata["title"]
date_modified = metadata["modified"]
year = date_modified.split('-')[0]
doi = metadata['distribution'][-1]['accessURL'].strip('https://dx.doi.org/')
url = metadata['distribution'][0]['downloadURL']
access_date = str(datetime.datetime.utcnow())

constants = requests.get(url).json()

output = '''
"""
This is a automatically generated file from the {0} NIST fundamental constants.
Title: {1}
Date: {2}
DOI: {3}
URL: {4}
Access Date: {5} UTC

File Authors: QCElemental Authors
"""


'''.format(year, title, date_modified, doi, url, access_date)

constants_json = {
    "title": title,
    "date": date_modified,
    "doi": doi,
    "url": url,
    "access_data": access_date,
    "constants": {}
}

for pc in constants['constant']:
    value = pc['Value'].strip()
    uncertainty = pc['Uncertainty']
    if uncertainty == '(exact)':
        value = value.replace('...', '')

    constants_json["constants"][pc["Quantity "].lower()] = {
        "quantity": pc["Quantity "],
        "unit": pc["Unit"],
        "value": value.replace(" ", ""),
        'uncertainty': uncertainty
    }
output += "nist_{}_codata = {}".format(year, constants_json)

output = FormatCode(output)

fn = "nist_{}_codata.py".format(year)
with open(fn, "w") as handle:
    handle.write(output[0])



"""
This file will generate a JSON blob usable by QCElemental for physical constants
"""

import os
import math
import json
import datetime
import requests
import pandas as pd
import black


table_url = "https://physics.nist.gov/cuu/Constants/Table/allascii.txt"

title = "NIST-CODATA Internationally Recommended 2018 Values of the Fundamental Physical Constants - SRD 121"  # edited year from https://catalog.data.gov/dataset/nist-codata-fundamental-physical-constants-srd-121
date_modified = "2019-05-20"  # https://pml.nist.gov/cuu/Constants/bibliography.html
data_through = "2018-12-31"  # https://pml.nist.gov/cuu/Constants/bibliography.html
year = data_through.split("-")[0]
doi = ""  # LAB: I can't find a thing
url = table_url
access_date = str(datetime.datetime.utcnow())

constants = requests.get(url).text
with open("localtable", "w") as fp:
    fp.write(constants)
    # Warning: colspecs="infer" or infer_nrows=400 still misses some ends of names
    data_also = pd.read_fwf("localtable", widths=[60, 25, 24, 20], skiprows=[0, 1, 2, 3, 4, 5, 6, 7, 8, 10])
os.unlink("localtable")

constants = data_also.to_dict("index")

output = f'''
"""
This is a automatically generated file from the {year} NIST fundamental constants.
Title: {title}
Date: {date_modified}
DOI: {doi}
URL: {url}
Access Date: {access_date} UTC

File Authors: QCElemental Authors
"""


'''

constants_json = {
    "title": title,
    "date": date_modified,
    "doi": doi,
    "url": url,
    "access_data": access_date,
    "constants": {},
}

for pc in constants.values():
    value = pc["Value"].strip()
    uncertainty = pc["Uncertainty"]
    if uncertainty == "(exact)":
        value = value.replace("...", "")

    constants_json["constants"][pc["Quantity"].lower()] = {
        "quantity": pc["Quantity"],
        "unit": str(pc["Unit"]).replace("nan", "").replace("^-1", "^{-1}").replace("^-2", "^{-2}").replace("^-3", "^{-3}").replace("^-4", "^{-4}").replace("_90", "_{90}"),
        "value": value.replace(" ", ""),
        "uncertainty": uncertainty,
    }
output += "nist_{}_codata = {}".format(year, constants_json)

output = black.format_str(output, mode=black.FileMode())

fn = f"nist_{year}_codata.py"
with open(fn, "w") as handle:
    handle.write(output)



"""
This file will generate a JSON blob usable by QCElemental for physical constants
"""

import json
import datetime
import requests
from yapf.yapflib.yapf_api import FormatCode
import re
from decimal import Decimal

# from https://www.nist.gov/pml/periodic-table-elements on 30 Aug 2018
#   NIST SP 966 (July 2018)
element_names = [
    'Hydrogen',
    'Helium',
    'Lithium',
    'Beryllium',
    'Boron',
    'Carbon',
    'Nitrogen',
    'Oxygen',
    'Fluorine',
    'Neon',
    'Sodium',
    'Magnesium',
    'Aluminum',
    'Silicon',
    'Phosphorus',
    'Sulfur',
    'Chlorine',
    'Argon',
    'Potassium',
    'Calcium',
    'Scandium',
    'Titanium',
    'Vanadium',
    'Chromium',
    'Manganese',
    'Iron',
    'Cobalt',
    'Nickel',
    'Copper',
    'Zinc',
    'Gallium',
    'Germanium',
    'Arsenic',
    'Selenium',
    'Bromine',
    'Krypton',
    'Rubidium',
    'Strontium',
    'Yttrium',
    'Zirconium',
    'Niobium',
    'Molybdenum',
    'Technetium',
    'Ruthenium',
    'Rhodium',
    'Palladium',
    'Silver',
    'Cadmium',
    'Indium',
    'Tin',
    'Antimony',
    'Tellurium',
    'Iodine',
    'Xenon',
    'Cesium',
    'Barium',
    'Lanthanum',
    'Cerium',
    'Praseodymium',
    'Neodymium',
    'Promethium',
    'Samarium',
    'Europium',
    'Gadolinium',
    'Terbium',
    'Dysprosium',
    'Holmium',
    'Erbium',
    'Thulium',
    'Ytterbium',
    'Lutetium',
    'Hafnium',
    'Tantalum',
    'Tungsten',
    'Rhenium',
    'Osmium',
    'Iridium',
    'Platinum',
    'Gold',
    'Mercury',
    'Thallium',
    'Lead',
    'Bismuth',
    'Polonium',
    'Astatine',
    'Radon',
    'Francium',
    'Radium',
    'Actinium',
    'Thorium',
    'Protactinium',
    'Uranium',
    'Neptunium',
    'Plutonium',
    'Americium',
    'Curium',
    'Berkelium',
    'Californium',
    'Einsteinium',
    'Fermium',
    'Mendelevium',
    'Nobelium',
    'Lawrencium',
    'Rutherfordium',
    'Dubnium',
    'Seaborgium',
    'Bohrium',
    'Hassium',
    'Meitnerium',
    'Darmstadtium',
    'Roentgenium',
    'Copernicium',
    'Nihonium',
    'Flerovium',
    'Moscovium',
    'Livermorium',
    'Tennessine',
    'Oganesson',
]

# from https://www.nist.gov/pml/periodic-table-elements on 30 Aug 2018
#   NIST SP 966 (July 2018)

longest_lived_isotope_for_unstable_elements = {
    'Tc': 98,
    'Pm': 145,
    'Po': 209,
    'At': 210,
    'Rn': 222,
    'Fr': 223,
    'Ra': 226,
    'Ac': 227,
    'Np': 237,
    'Pu': 244,
    'Am': 243,
    'Cm': 247,
    'Bk': 247,
    'Cf': 251,
    'Es': 252,
    'Fm': 257,
    'Md': 258,
    'No': 259,
    'Lr': 266,
    'Rf': 267,
    'Db': 268,
    'Sg': 271,
    'Bh': 270,
    'Hs': 269,
    'Mt': 278,
    'Ds': 281,
    'Rg': 282,
    'Cn': 285,
    'Nh': 286,
    'Fl': 289,
    'Mc': 289,
    'Lv': 293,
    'Ts': 294,
    'Og': 294,
}

data_url = "https://nist.gov/srd/srd_data//srd144_Atomic_Weights_and_Isotopic_Compositions_for_All_Elements.json"

title = "Atomic Weights and Isotopic Compositions with Relative Atomic Masses - SRD144"
date_modified = "2011-01-14"
year = date_modified.split('-')[0]
doi = "10.1351/PAC-REP-10-06-02"
url = data_url
access_date = str(datetime.datetime.utcnow())

atomic_weights_data = requests.get(url).json()

output = '''
"""
This is a automatically generated file from the {0} NIST atomic weights.
Title: {1}
Date: {2}
DOI: {3}
URL: {4}
Access Date: {5} UTC

File Authors: QCElemental Authors
"""


'''.format(year, title, date_modified, doi, url, access_date)

atomic_weights_json = {"title": title, "date": date_modified, "doi": doi, "url": url, "access_data": access_date}

# length number of elements
Z = [0]  # , 1, 2, ...
E = ['X']  # , H, He, ...
name = ['Dummy']  # , Hydrogen, Helium, ...

# length number of elements plus number of isotopes
_EE = ['X', 'X']  # , H, H, H, ..., He, He, He, ...
EA = ['X', 'X0']  # , H, H1, H2, ..., He, He3, He4, ...
A = [0, 0]  # , 1, 1, 2, ..., 4, 3, 4, ...
masses = ["0", "0"]  # , 1.0078, 1.0078, 2.014, ..., 4.0026, 3.016, 4.0026, ...V

uncertain_value = re.compile(r"""(?P<value>[\d.]+)(?P<uncertainty>\([\d#]+\))?""")
aliases = {'D': 'H2', 'T': 'H3'}

newnames = {'Uut': 'Nh', 'Uup': 'Mc', 'Uus': 'Ts'}
for delem in atomic_weights_data['data']:
    symbol = delem['Atomic Symbol']
    delem['Atomic Symbol'] = newnames.get(symbol, symbol)
    for diso in delem['isotopes']:
        symbol = diso['Atomic Symbol']
        diso['Atomic Symbol'] = newnames.get(symbol, symbol)

# element loop
for delem in atomic_weights_data['data']:
    mass_of_most_common_isotope = None
    mass_number_of_most_common_isotope = None
    max_isotopic_contribution = 0.0

    # isotope loop
    for diso in delem['isotopes']:
        mobj = re.match(uncertain_value, diso['Relative Atomic Mass'])

        if mobj:
            mass = mobj.group('value')
        else:
            raise ValueError('Trouble parsing mass string ({}) for element ({})'.format(
                diso['Relative Atomic Mass'], diso['Atomic Symbol']))

        a = int(diso['Mass Number'])

        if diso['Atomic Symbol'] in aliases:
            _EE.append('H')
            EA.append(aliases[diso['Atomic Symbol']])
            A.append(a)
            masses.append(mass)

            _EE.append('H')
            EA.append(diso['Atomic Symbol'])
            A.append(a)
            masses.append(mass)

        else:
            _EE.append(diso['Atomic Symbol'])
            EA.append(diso['Atomic Symbol'] + diso['Mass Number'])
            A.append(a)
            masses.append(mass)

        if 'Isotopic Composition' in diso:
            mobj = re.match(uncertain_value, diso['Isotopic Composition'])

            if mobj:
                if float(mobj.group('value')) > max_isotopic_contribution:
                    mass_of_most_common_isotope = mass
                    mass_number_of_most_common_isotope = a
                    max_isotopic_contribution = float(mobj.group('value'))

    # Source atomic_weights_and_isotopic_compositions_for_all_elements deals with isotopic composition of
    #   stable elements. For unstable elements, need another source for the longest-lived isotope.
    if mass_of_most_common_isotope is None:
        mass_number_of_most_common_isotope = longest_lived_isotope_for_unstable_elements[diso['Atomic Symbol']]
        eliso = delem['Atomic Symbol'] + str(mass_number_of_most_common_isotope)
        mass_of_most_common_isotope = masses[EA.index(eliso)]

    _EE.append(delem['Atomic Symbol'])
    EA.append(delem['Atomic Symbol'])
    A.append(mass_number_of_most_common_isotope)
    masses.append(mass_of_most_common_isotope)

    z = int(delem['Atomic Number'])

    Z.append(z)
    E.append(delem['Atomic Symbol'])
    name.append(element_names[z - 1].capitalize())

atomic_weights_json["Z"] = Z
atomic_weights_json["E"] = E
atomic_weights_json["name"] = name
atomic_weights_json["_EE"] = _EE
atomic_weights_json["EA"] = EA
atomic_weights_json["A"] = A
atomic_weights_json["mass"] = masses
output += "nist_{}_atomic_weights = {}".format(year, json.dumps(atomic_weights_json))

# output = FormatCode(output)[0]
fn = "nist_{}_atomic_weights.py".format(year)
with open(fn, "w") as handle:
    handle.write(output)
